#include "btree_gist.h"

#define BTREE_GIST_FLOAT4 1

typedef struct float4key
{
	float4		lower;
	float4		upper;
}	float4KEY;


/*
** float4key in/out
*/
PG_FUNCTION_INFO_V1(float4key_in);
PG_FUNCTION_INFO_V1(float4key_out);
Datum		float4key_in(PG_FUNCTION_ARGS);
Datum		float4key_out(PG_FUNCTION_ARGS);

/*
** float4 ops
*/
PG_FUNCTION_INFO_V1(gfloat4_compress);
PG_FUNCTION_INFO_V1(gfloat4_union);
PG_FUNCTION_INFO_V1(gfloat4_picksplit);
PG_FUNCTION_INFO_V1(gfloat4_consistent);
PG_FUNCTION_INFO_V1(gfloat4_penalty);
PG_FUNCTION_INFO_V1(gfloat4_same);

Datum		gfloat4_compress(PG_FUNCTION_ARGS);
Datum		gfloat4_union(PG_FUNCTION_ARGS);
Datum		gfloat4_picksplit(PG_FUNCTION_ARGS);
Datum		gfloat4_consistent(PG_FUNCTION_ARGS);
Datum		gfloat4_penalty(PG_FUNCTION_ARGS);
Datum		gfloat4_same(PG_FUNCTION_ARGS);

static void gfloat4_binary_union(Datum *r1, char *r2);
static int	float4key_cmp(const void *a, const void *b);


/**************************************************
 * float4 ops
 **************************************************/

Datum
gfloat4_compress(PG_FUNCTION_ARGS)
{
	GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
	GISTENTRY  *retval;

	if (entry->leafkey)
	{
		float4KEY    *r = ( float4KEY * ) palloc(sizeof(float4KEY));
    #ifdef BTREE_GIST_INT2
    	int16		leaf = DatumGetInt16(entry->key);
    #endif
    #ifdef BTREE_GIST_INT4
    	int32		leaf = DatumGetInt32(entry->key);
    #endif
    #ifdef BTREE_GIST_INT8
    	int64		leaf = DatumGetInt64(entry->key);
    #endif
    #ifdef BTREE_GIST_FLOAT4
    	float4	leaf = DatumGetFloat4(entry->key);
    #endif
    #ifdef BTREE_GIST_FLOAT8
    	float8	leaf = DatumGetFloat8(entry->key);
    #endif

		retval = palloc(sizeof(GISTENTRY));
		r->lower = r->upper = leaf ;

		gistentryinit(*retval, PointerGetDatum(r), entry->rel, entry->page,
					  entry->offset, sizeof(float4KEY), FALSE);

	}
	else
		retval = entry;
	PG_RETURN_POINTER(retval);
}

Datum
gfloat4_consistent(PG_FUNCTION_ARGS)
{
	GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
  #ifdef BTREE_GIST_INT2
  	int16		query = PG_GETARG_INT16(1);
  #endif
  #ifdef BTREE_GIST_INT4
  	int32		query = PG_GETARG_INT32(1);
  #endif
  #ifdef BTREE_GIST_INT8
  	int64		query = PG_GETARG_INT64(1);
  #endif
  #ifdef BTREE_GIST_FLOAT4
  	float4		query = PG_GETARG_FLOAT4(1);
  #endif
  #ifdef BTREE_GIST_FLOAT8
  	float8		query = PG_GETARG_FLOAT8(1);
  #endif
	float4KEY    *kkk = (float4KEY *) DatumGetPointer(entry->key);
	StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
	bool		retval;

	switch (strategy)
	{
		case BTLessEqualStrategyNumber:
			retval = (query >= kkk->lower);
			break;
		case BTLessStrategyNumber:
			if (GIST_LEAF(entry))
				retval = (query > kkk->lower);
			else
				retval = (query >= kkk->lower);
			break;
		case BTEqualStrategyNumber:
			/* in leaf page kkk->lower always = kkk->upper */
			if (GIST_LEAF(entry))
				retval = (query == kkk->lower);
			else
				retval = (kkk->lower <= query && query <= kkk->upper);
			break;
		case BTGreaterStrategyNumber:
			if (GIST_LEAF(entry))
				retval = (query < kkk->upper);
			else
				retval = (query <= kkk->upper);
			break;
		case BTGreaterEqualStrategyNumber:
			retval = (query <= kkk->upper);
			break;
		default:
			retval = FALSE;
	}
	PG_RETURN_BOOL(retval);
}

Datum
gfloat4_union(PG_FUNCTION_ARGS)
{
	bytea	   *entryvec = (bytea *) PG_GETARG_POINTER(0);
	int			i,
				numranges;
	float4KEY    *cur,
			   *out = palloc(sizeof(float4KEY));

	numranges = (VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY);
	*(int *) PG_GETARG_POINTER(1) = sizeof(float4KEY);

	cur = (float4KEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[0].key));
	out->lower = cur->lower;
	out->upper = cur->upper;

	for (i = 1; i < numranges; i++)
	{
		cur = (float4KEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[i].key));
		if (out->lower > cur->lower)
			out->lower = cur->lower;
		if (out->upper < cur->upper)
			out->upper = cur->upper;
	}

	PG_RETURN_POINTER(out);
}

Datum
gfloat4_penalty(PG_FUNCTION_ARGS)
{
	float4KEY    *origentry = (float4KEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(0))->key);
	float4KEY    *newentry = (float4KEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(1))->key);
	float	   *result = (float *) PG_GETARG_POINTER(2);

	*result = Max(newentry->upper - origentry->upper, 0) +
		Max(origentry->lower - newentry->lower, 0);

	PG_RETURN_POINTER(result);
}

Datum
gfloat4_picksplit(PG_FUNCTION_ARGS)
{
	PG_RETURN_POINTER(btree_picksplit(
									  (bytea *) PG_GETARG_POINTER(0),
								  (GIST_SPLITVEC *) PG_GETARG_POINTER(1),
									  gfloat4_binary_union,
									  float4key_cmp
									  ));
}

Datum
gfloat4_same(PG_FUNCTION_ARGS)
{
	float4KEY    *b1 = (float4KEY *) PG_GETARG_POINTER(0);
	float4KEY    *b2 = (float4KEY *) PG_GETARG_POINTER(1);
	bool	   *result = (bool *) PG_GETARG_POINTER(2);

	*result = (b1->lower == b2->lower && b1->upper == b2->upper) ? TRUE : FALSE;
	PG_RETURN_POINTER(result);
}

static void
gfloat4_binary_union(Datum *r1, char *r2)
{
	float4KEY    *b1;
	float4KEY    *b2 = (float4KEY *) r2;

	if (!DatumGetPointer(*r1))
	{
		*r1 = PointerGetDatum(palloc(sizeof(float4KEY)));
		b1 = (float4KEY *) DatumGetPointer(*r1);
		b1->upper = b2->upper;
		b1->lower = b2->lower;
	}
	else
	{
		b1 = (float4KEY *) DatumGetPointer(*r1);

		b1->lower = (b1->lower > b2->lower) ?
			b2->lower : b1->lower;
		b1->upper = (b1->upper > b2->upper) ?
			b1->upper : b2->upper;
	}

}


static int
float4key_cmp(const void *a, const void *b)
{


	if (((float4KEY *) (((RIX *) a)->r))->lower > ((float4KEY *) (((RIX *) b)->r))->lower){
    return 1;
	} else if (((float4KEY *) (((RIX *) a)->r))->lower < ((float4KEY *) (((RIX *) b)->r))->lower){
    return -1;
  } else {
    return 0;
  }

}


/**************************************************
 * In/Out for keys
 **************************************************/

Datum
float4key_in(PG_FUNCTION_ARGS)
{
  ereport(ERROR,
		(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
		 errmsg("<datatype>key_in() not implemented")));

  PG_RETURN_POINTER(NULL);
}

Datum
float4key_out(PG_FUNCTION_ARGS)
{
  ereport(ERROR,
		(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
		 errmsg("<datatype>key_out() not implemented")));

  PG_RETURN_POINTER(NULL);
}

