#include "btree_gist.h"

#define BTREE_GIST_INT8 1

typedef struct int64key
{
	int64		lower;
	int64		upper;
}	int64KEY;


/*
** int64key in/out
*/
PG_FUNCTION_INFO_V1(int8key_in);
PG_FUNCTION_INFO_V1(int8key_out);
Datum		int8key_in(PG_FUNCTION_ARGS);
Datum		int8key_out(PG_FUNCTION_ARGS);

/*
** int64 ops
*/
PG_FUNCTION_INFO_V1(gint8_compress);
PG_FUNCTION_INFO_V1(gint8_union);
PG_FUNCTION_INFO_V1(gint8_picksplit);
PG_FUNCTION_INFO_V1(gint8_consistent);
PG_FUNCTION_INFO_V1(gint8_penalty);
PG_FUNCTION_INFO_V1(gint8_same);

Datum		gint8_compress(PG_FUNCTION_ARGS);
Datum		gint8_union(PG_FUNCTION_ARGS);
Datum		gint8_picksplit(PG_FUNCTION_ARGS);
Datum		gint8_consistent(PG_FUNCTION_ARGS);
Datum		gint8_penalty(PG_FUNCTION_ARGS);
Datum		gint8_same(PG_FUNCTION_ARGS);

static void gint8_binary_union(Datum *r1, char *r2);
static int	int8key_cmp(const void *a, const void *b);


/**************************************************
 * int64 ops
 **************************************************/

Datum
gint8_compress(PG_FUNCTION_ARGS)
{
	GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
	GISTENTRY  *retval;

	if (entry->leafkey)
	{
		int64KEY    *r = ( int64KEY * ) palloc(sizeof(int64KEY));
    #ifdef BTREE_GIST_INT2
    	int16		leaf = DatumGetInt16(entry->key);
    #endif
    #ifdef BTREE_GIST_INT4
    	int32		leaf = DatumGetInt32(entry->key);
    #endif
    #ifdef BTREE_GIST_INT8
    	int64		leaf = DatumGetInt64(entry->key);
    #endif
    #ifdef BTREE_GIST_FLOAT4
    	float4	leaf = DatumGetFloat4(entry->key);
    #endif
    #ifdef BTREE_GIST_FLOAT8
    	float8	leaf = DatumGetFloat8(entry->key);
    #endif

		retval = palloc(sizeof(GISTENTRY));
		r->lower = r->upper = leaf ;

		gistentryinit(*retval, PointerGetDatum(r), entry->rel, entry->page,
					  entry->offset, sizeof(int64KEY), FALSE);

	}
	else
		retval = entry;
	PG_RETURN_POINTER(retval);
}

Datum
gint8_consistent(PG_FUNCTION_ARGS)
{
	GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
  #ifdef BTREE_GIST_INT2
  	int16		query = PG_GETARG_INT16(1);
  #endif
  #ifdef BTREE_GIST_INT4
  	int32		query = PG_GETARG_INT32(1);
  #endif
  #ifdef BTREE_GIST_INT8
  	int64		query = PG_GETARG_INT64(1);
  #endif
  #ifdef BTREE_GIST_FLOAT4
  	float4		query = PG_GETARG_FLOAT4(1);
  #endif
  #ifdef BTREE_GIST_FLOAT8
  	float8		query = PG_GETARG_FLOAT8(1);
  #endif
	int64KEY    *kkk = (int64KEY *) DatumGetPointer(entry->key);
	StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
	bool		retval;

	switch (strategy)
	{
		case BTLessEqualStrategyNumber:
			retval = (query >= kkk->lower);
			break;
		case BTLessStrategyNumber:
			if (GIST_LEAF(entry))
				retval = (query > kkk->lower);
			else
				retval = (query >= kkk->lower);
			break;
		case BTEqualStrategyNumber:
			/* in leaf page kkk->lower always = kkk->upper */
			if (GIST_LEAF(entry))
				retval = (query == kkk->lower);
			else
				retval = (kkk->lower <= query && query <= kkk->upper);
			break;
		case BTGreaterStrategyNumber:
			if (GIST_LEAF(entry))
				retval = (query < kkk->upper);
			else
				retval = (query <= kkk->upper);
			break;
		case BTGreaterEqualStrategyNumber:
			retval = (query <= kkk->upper);
			break;
		default:
			retval = FALSE;
	}
	PG_RETURN_BOOL(retval);
}

Datum
gint8_union(PG_FUNCTION_ARGS)
{
	bytea	   *entryvec = (bytea *) PG_GETARG_POINTER(0);
	int			i,
				numranges;
	int64KEY    *cur,
			   *out = palloc(sizeof(int64KEY));

	numranges = (VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY);
	*(int *) PG_GETARG_POINTER(1) = sizeof(int64KEY);

	cur = (int64KEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[0].key));
	out->lower = cur->lower;
	out->upper = cur->upper;

	for (i = 1; i < numranges; i++)
	{
		cur = (int64KEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[i].key));
		if (out->lower > cur->lower)
			out->lower = cur->lower;
		if (out->upper < cur->upper)
			out->upper = cur->upper;
	}

	PG_RETURN_POINTER(out);
}

Datum
gint8_penalty(PG_FUNCTION_ARGS)
{
	int64KEY    *origentry = (int64KEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(0))->key);
	int64KEY    *newentry = (int64KEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(1))->key);
	float	   *result = (float *) PG_GETARG_POINTER(2);

	*result = Max(newentry->upper - origentry->upper, 0) +
		Max(origentry->lower - newentry->lower, 0);

	PG_RETURN_POINTER(result);
}

Datum
gint8_picksplit(PG_FUNCTION_ARGS)
{
	PG_RETURN_POINTER(btree_picksplit(
									  (bytea *) PG_GETARG_POINTER(0),
								  (GIST_SPLITVEC *) PG_GETARG_POINTER(1),
									  gint8_binary_union,
									  int8key_cmp
									  ));
}

Datum
gint8_same(PG_FUNCTION_ARGS)
{
	int64KEY    *b1 = (int64KEY *) PG_GETARG_POINTER(0);
	int64KEY    *b2 = (int64KEY *) PG_GETARG_POINTER(1);
	bool	   *result = (bool *) PG_GETARG_POINTER(2);

	*result = (b1->lower == b2->lower && b1->upper == b2->upper) ? TRUE : FALSE;
	PG_RETURN_POINTER(result);
}

static void
gint8_binary_union(Datum *r1, char *r2)
{
	int64KEY    *b1;
	int64KEY    *b2 = (int64KEY *) r2;

	if (!DatumGetPointer(*r1))
	{
		*r1 = PointerGetDatum(palloc(sizeof(int64KEY)));
		b1 = (int64KEY *) DatumGetPointer(*r1);
		b1->upper = b2->upper;
		b1->lower = b2->lower;
	}
	else
	{
		b1 = (int64KEY *) DatumGetPointer(*r1);

		b1->lower = (b1->lower > b2->lower) ?
			b2->lower : b1->lower;
		b1->upper = (b1->upper > b2->upper) ?
			b1->upper : b2->upper;
	}

}


static int
int8key_cmp(const void *a, const void *b)
{


	if (((int64KEY *) (((RIX *) a)->r))->lower > ((int64KEY *) (((RIX *) b)->r))->lower){
    return 1;
	} else if (((int64KEY *) (((RIX *) a)->r))->lower < ((int64KEY *) (((RIX *) b)->r))->lower){
    return -1;
  } else {
    return 0;
  }

}


/**************************************************
 * In/Out for keys
 **************************************************/

Datum
int8key_in(PG_FUNCTION_ARGS)
{
  ereport(ERROR,
		(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
		 errmsg("<datatype>key_in() not implemented")));

  PG_RETURN_POINTER(NULL);
}

Datum
int8key_out(PG_FUNCTION_ARGS)
{
  ereport(ERROR,
		(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
		 errmsg("<datatype>key_out() not implemented")));

  PG_RETURN_POINTER(NULL);
}

